REBOL [
	Title: "Admin protocol"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Date: 03/10/2004
	Version: 1.0.0
	Purpose: "UniServe's remote management protocol"
]

install-protocol [
	name: 'Admin
    port-id: 9800
	
	make-session-class: does [
		context [
			state: 'header
			last-cmd: none
		]
	]
	
	validate?: func [data][
		if not find [start stop list stats] server/user-data/last-cmd: data/1 [
			server/user-data/last-cmd: none
		]
		true
	]
	
	on-init-port: does [stop-at: 4]
	
	on-connected: does [
		server/user-data: make-session-class
		on-logged
	]
	
	on-received: func [data /local su][
		su: server/user-data
		switch su/state [
			header [
				stop-at: to integer! data
				su/state: 'body
			]
			body [		
				data: first load/all to-string data			
				either data/1 = 'ok [
					on-response switch/default su/last-cmd [
							list  [sort/skip data/2 4]
							stats [data/2]
						][first reduce at data 2]
						su/last-cmd
				][
					on-error data/2
				]
				su/state: 'header
				stop-at: 4		; ready for next request
			]
		]
	]
	
	new-insert-port: func [port blk [block!]][
		either validate? blk [
			blk: mold blk
			write-server debase/base to-hex length? blk 16
			write-server blk
		][
			on-error 'invalid-command
		]
	]

	on-error: func [data][
		log/error mold data
	]

	events: [
		on-logged			; []
		on-response 		; [msg last-cmd]
		on-error			; [code]
		on-close-server		; []
	]
]