REBOL [
	Title: "POP3 Async Protocol"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 0.9.0
	Date: 02/10/2004
	Comment:  {
		Warning: the "Delete" mode has not been tested, so its disabled for now !
	
		POP Command Dialect (BNF syntax) :
	
		commands = get-cmd | del-cmd

		get-cmd = 'get ['all | n | 'last [n] ['lines n | 'head | 'size]]
		del-cmd = 'delete [n | 'all]
		n = {integer! value}
	}
]

install-protocol [
	name: 'POP
	port-id: 110
	verbose: 0
	
	multi-line-end: rejoin [crlf #"." crlf]
	
	stop-at: crlf
	
	send: func [req][
		if verbose > 0 [log/info rejoin ["request >> " req]]
		write-server req
	]
	
	answer-ok?: func [data su][
		su/last-answer: trim/tail to-string data
		either find/part data "+OK" 3 [true][
			on-cmd-failed server su/state su/last-answer
			false
		]
	]
	
	on-connected: has [su][
		su: server/user-data: context [
			state: 'start
			user: pass:	last-answer: last-cmd: id-list: none
			nb: total: id: 0
			msgs: make block! 128
		]
		su/user: server/user
		su/pass: server/pass
		on-login-start server su
	]
	
	on-received: func [data /local su][	
		su: server/user-data
		switch su/state [
			start [
				if answer-ok? data su [
					send rejoin ["USER " su/user crlf]
					su/state: 'user
				]
			]
			user [
				if answer-ok? data su [
					send rejoin ["PASS " su/pass crlf]
					su/state: 'pass
				]
			]
			pass [
				if answer-ok? data su [
					send rejoin ["STAT" crlf]
					su/state: 'stats
				]
			]
			stats [
				if answer-ok? data su [
					data: parse skip data 4 ""
					su/nb: to integer! data/1
					su/total: to integer! data/2
					
					on-login-ok server su
				]
				
			]
			get-start [
				if answer-ok? data su [
					stop-at: multi-line-end
					su/state: 'get-end
				]
			]
			get-end [
				on-message server to-string data su/id
				su/state: none
				stop-at: crlf
				if su/id-list [
					change su/id-list su/id: su/id + 1
					either su/id-list/1 = su/id-list/2 [
						su/id-list: none
						on-all-messages server
					][
						poke su/last-cmd 2 su/id
						send reform su/last-cmd
						su/state: 'get-start
					]
				]
			]
			delete [
				either answer-ok? data su [
					on-deleted server su/id yes
					if su/id-list [
						change su/id-list su/id: su/id + 1
						either su/id-list/1 = su/id-list/2 [
							su/id-list: none
							on-all-deleted server
						][
							poke su/last-cmd 2 su/id
							send reform su/last-cmd
						]
					]	
				][
					on-deleted server su/id no
				]
			]
			size-start [
				if answer-ok? data su [
					either su/id-list [
						stop-at: multi-line-end
						su/state: 'size-end
					][
						on-msg-size server to-integer second parse skip data 4 ""
						su/state: none
					]
				]
			]
			size-end [
				data: load/all to-string data
				remove back tail data
				on-msg-size server data
				su/id-list: none
				stop-at: crlf
			]
		][
			log/error reform ["unknown state" mold su/state]
		]
	]
	
	new-insert-port: func [
		port spec [block!]
		/local su out valid target lines
	][
		su: server/user-data		
		either valid: parse spec [
			'get (out: ["RETR"]	su/state: 'get-start) [
				set target integer! (insert tail out target) 
				| 'all (out: ["RETR" 1] su/id-list: reduce [1 su/nb])
			] 
			opt [
				'lines set lines integer! (
					out: ["TOP" su/id: any [target 1] lines]
				)
				| 'head (out: ["TOP" su/id: any [target 1] "0"])
				| 'size (
					out: ["LIST" any [su/id: target ""]]
					su/state: 'size-start
				)
			]
			| 'delete (out: ["DELE"] su/state: 'delete)
			   set target [
				   integer! (insert tail out su/id: target)
				   | block! (
				   	   insert tail out su/id: target/1
				   	   su/id-list: target
				   )
			   ]
			   | 'all (
			   	   out: ["DELE" su/id: 1]
			   	   su/id-list: reduce [1 su/nb]
			   )			
		][
			send reform head insert tail su/last-cmd: out crlf
		][
			on-error 'cmd-syntax
		]
	]
		
	events: [
		on-login-start	; [port [port!] auth [object!]]
		on-login-ok		; [port [port!] mbox [object!]]
		on-message		; [port [port!] msg [string!] id [integer!]]
		on-all-messages	; [port [port!]]
		on-msg-size		; [port [port!] size [integer! block!]]
		on-deleted		; [port [port!] id [integer!] result [logic!]]
		on-all-deleted	; [port [port!]]
		on-cmd-failed	; [port [port!] cmd [word!] err-msg [string!]]
	]
]



