REBOL [
	Title: "Uniserve: task-master service"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 1.0.0
	Date: 28/09/2004
]

install-service [
	name: 'task-master
	port-id: 9799
	;hidden: yes
	verbose: 0
	
	pool-list: make block! 50
	obj: none
	queue: copy []

	share [
		pool-start: 0	; number of helper process when Uniserve starts
		pool-max: 	4	; maximum number of helper process
		pool-count: 0	; read-only - do not change it !!
		job-max: 	100	; max size of the waiting jobs queue
		do-task: func [data service][process-task data service service/client]
	]

	task: context [
		busy: no
		ctx: none
	]
	
	fork: does [
		launch either system/version/4 = 3 [
			{"%services/task-master/task-handler.r"}		; use "-qsw --script" for true bg processes
		][
			"%services/task-master/task-handler.r"			; -qws not working !!!
		]
		shared/pool-count: shared/pool-count + 1
	]	
	
	send-data: func [port data][
		write-client/with debase/base to-hex length? data: mold data 16 port
		write-client/with data port
	]
	
	process-task: func [data server port /local list obj][
		list: :pool-list
		while [all [not tail? list list/1/user-data/busy]][list: next list]
		either tail? list [
			if shared/pool-max > shared/pool-count [
				fork
				if verbose > 0 [log/info "launching new process"]
			]
			either shared/job-max > length? queue  [
				repend/only queue [copy/deep data server port]
			][
				if in obj: client/user-data/ctx/1 'on-task-failed [
					obj/on-task-failed 'overload
				]
			]
		][
			if verbose > 0 [log/info "new task affected"]
			send-data list/1 reduce [server/name server/module data]
			list/1/user-data/busy: yes
			list/1/user-data/ctx: reduce [server port]
		]
	]
	
	on-load: does [if integer? shared/pool-start [loop shared/pool-start [fork]]]

	on-new-client: has [tmp][
		if client/remote-ip <> 127.0.0.1 [close-client exit]
		client/timeout: 15
		client/user-data: make task []
		append pool-list :client
		stop-at: 4
		if verbose > 0 [log/info "new slave process connected"]
		if not empty? queue [
			tmp: queue/1
			send-data client reduce [tmp/2/name tmp/2/module tmp/1]
			client/user-data/busy: yes
			client/user-data/ctx: reduce [tmp/2 tmp/3]
			remove queue
			if verbose > 0 [log/info "new slave got job"]
		]
	]
	
	on-close-client: does [
		remove find pool-list :client
		if client/user-data/busy [
			if in obj: client/user-data/ctx/1 'on-task-failed [
				obj/on-task-failed 'error
			]
		]
		if verbose > 0 [log/info "slave process closed"]
		shared/pool-count: shared/pool-count - 1
	]
	
	
	on-received: func [data /local tmp][
		either stop-at = 4 [
			if verbose > 1 [log/info "header received"]
			stop-at: to integer! copy data
		][
			if verbose > 1 [log/info "body received"]
			client/user-data/ctx/1/client: client/user-data/ctx/2
			data: load to string! data
			either data/1 = 'ok [
				client/user-data/ctx/1/on-task-done data/2
			][
				if in obj: client/user-data/ctx/1 'on-task-failed [obj/on-task-failed data/2]
			]
			either empty? queue [
				client/user-data/busy: no
				client/user-data/ctx: none
			][
				tmp: queue/1
				send-data client reduce [tmp/2/name tmp/2/module tmp/1]
				client/user-data/ctx: reduce [tmp/2 tmp/3]
				remove queue
				if verbose > 0 [log/info "no wait-state slave reuse"]
			]
			stop-at: 4
		]
	]
]