REBOL [
	Title: "UniServe kernel"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Copyright: "@ 2002-2004 SOFTINNOV"
	Email: nr@softinnov.com
	Date: 03/10/2004
	File: %uni-engine.r
	Version: 0.9.9
	Purpose: "Multi-protocol asynchrone client/server framework"
	License: {
		BSD License, read the complete text in %docs/license.txt
		This license covers this source code and all files in this
		archive.
	}
	History: [
		0.9.10	11/10/2004 {
			o Bugfix in 'insert-port (added correct path to 'write-peer)
		}
		0.9.9	03/10/2004 "History starts"
	]
]

do %libs/log.r

; --- 'launch function rewrited using 'call ---
launch: func [cmd [string!]][
	call join form to-local-file system/options/boot [" -s " cmd]
]

; --- patch for 'in function to accept port! datatype
if not find ti: third third :in port! [append ti port!]
; ---

uniserve: make log-class [
	name: 'uniserve
	verbose: 0		; > 0 => show debug info

	services:  make block! 20
	protocols: make block! 20
	
	services-path:  %services
	protocols-path: %protocols 

	ip-buffer: make binary! ip-buf-size: 64 * 1024
	
	spwl: system/ports/wait-list

	plugin-class: context [
		name: port-id: hidden: peer: stop-at: shared: module: events:
		on-received: on-raw-received: set-peer: on-error: none
		scheme: 'tcp

		write-peer: func [data [binary! string! file!] /with port][
			if not with [port: peer]
			if not find spwl port [
				if verbose > 0 [log/warn "writing aborted: premature connection close"]
				return true
			]
			insert tail port/locals/write-queue :data
			port/async-modes: [read write]
			if verbose > 2 [log/info ["output =>^/" mold :data]]
			true
		]
		close-peer: func [/force /with port][
			if not with [port: peer]
			either any [force empty? port/locals/write-queue][
				uniserve/close-connection port
			][
				port/locals/flag-close: on
			]
		]
		share: func [spec [block!]][uniserve/shared: make uniserve/shared spec]
	]

	service-class: make plugin-class [
		write-client: :write-peer
		close-client: :close-peer
		set-peer: func [port][client: peer: port]
		on-new-client: on-close-client: client: none
	]

	protocol-class: make plugin-class [
		write-server: func [data /with port][
			if not with [port: server]
			write-peer/with data port
		]
		close-server: func [/with port][
			if not with [port: server]
			close-peer/with port
		]
		set-peer: func [port][server: peer: port]
		;on-connect: actions: on-init: on-closed:
		on-init-port: on-close-server: events: service: server: 
		on-connected: on-failed: none
		connect-retries: 3	
	]

	set 'open-port func [
		url [url! string!] events [block!] /with vars [block!] 
		/local port proto scheme host target usr psw px-host px-port 
	][
		forall events [
			if set-word? events/1 [change events to-lit-word events/1]
		]
		parse/all url [
			copy scheme to "://" 3 skip 
			copy host [to slash | to end]
			[end | copy target to end]
		]
		if pos: find host #"@" [
			 set [usr psw] parse copy/part host pos ":"
			 pos: next pos
		]
		if pos: find host: any [pos host] #":" [
			host: copy/part host pos
			port-id: to-integer copy next pos 
		]	
		scheme: to word! scheme
		if scheme = 'dns [
			return open-dns host reduce [scheme head events vars] :on-dns-data
		]
		port: make port! append [
			scheme: 'tcp
			user: usr
			pass: psw
		] any [all [with vars] []]
		port/host: any [
			all [px-host: system/schemes/default/proxy/host px-host] host
		]
		attempt [port/host: to-tuple port/host]
		if px-port: system/schemes/default/proxy/port-id [
			port/port-id: px-port
		]
		port/target: any [target "/"]		
		if not proto: select protocols scheme [
			make error! join "unknown protocol " scheme
		]
		port/locals: reduce [scheme proto/connect-retries]
		port/scheme: proto/scheme
		if any [not port/port-id zero? port/port-id][
			port/port-id: proto/port-id
		]		
		port/url: url
		proto/on-init-port port url	
		port/state/flags: system/standard/port-flags/pass-thru
		port/state/func: 3				; black magic !
		port/user-data: head events
		either tuple? port/host [
			append spwl port
			if verbose > 1 [log/info reform ["Connecting to" port/host ":" port/port-id]]
			port/async-modes: 'connect
			port/awake: :on-connect
			port/awake port		; try to connect at once			
		][
			open-dns port/host port :on-resolve
		]
		port
	]

	set 'insert-port func [port data /local name][
		either in port/locals/handler name: 'new-insert-port [
			fire-event/arg/arg2 port name port data
		][
			port/locals/handler/write-peer/with data port
		]
	]
	

	set 'close-port func [port][
		either all [
			object? port/locals
			in port/locals 'handler
		][
			port/locals/handler/close-peer/with port
		][
			uniserve/close-connection port
		]
	]
	
	open-service: func [name [word!] id /local model][
		either model: select services name [
			if error? try [
				append spwl open/binary/direct/no-wait
					make port! compose [
						scheme: (to-lit-word model/scheme)
						server-type: (model)
						port-id: (any [id model/port-id])
						async-modes: 'accept
						awake: :on-accept
					]
			][
				log/error [
					"cannot open server " mold name " on port "
					any [id model/port-id]
				]
				none
			]
		][
			log/error ["service " mold name " not installed!"]
			none
		]
	]

	find-service: func [name id /local list port][
	 	list: spwl
		forall list [
			if all [
				port? port: list/1
				in port 'server-type
				port/server-type/name = name
				port/port-id = any [id port/server-type/port-id]
			][
				return list
			]
		]
		false
	]
	
	close-service: func [name id /local pos][
		either pos: find-service name id [
			close pos/1
			remove pos
		][
			log/warn [mold name " not running!"]
			none
		]
	]
	
	init-connection: func [new /service server /local proto evt][
		new/locals: context [
			handler: any [
				all [service server/server-type]
				proto: select protocols new/locals/1
			]
			write-queue: copy []
			file-chunk: 64 * 1024
			file:
			flag-close: none
			stop: handler/stop-at
			in-buffer: make binary! 64 * 1024
			events: all [proto extract/index evt: reduce new/user-data 2 2]
			event-names: all [proto bind extract evt 2 in proto 'self]
		]
	]

	close-connection: func [port /bypass /local pos ctx][
		if all [
			not bypass
			object? port/locals
			in port/locals 'handler
		][
			ctx: port/locals/handler
			fire-event port any [
				in ctx 'on-close-client
				in ctx 'on-close-server
			]
		]
		remove pos: find spwl port
		if pos [close port]
		if verbose > 0 [log/info ["port closed : " port/remote-ip]]
	]

	remove-from-queue: func [list][
		if port? first list [close first list]
		remove list
	]

	fire-event: func [port name /arg value /arg2 value2 /local ctx pl err][
		if verbose > 1 [
			log/info rejoin [
				"Calling >" name "<" any [
					all [
						arg 
						series? value
						join " with " mold to-string copy/part value 50
					] ""
				]
			]
		]
		pl: port/locals
		ctx: pl/handler
		ctx/stop-at: pl/stop	
		if ctx/events [set pl/event-names pl/events]
		ctx/set-peer port
		if error? err: try [ctx/:name value value2][
			log/error rejoin [name " call failed with error: " mold disarm err]
			close-connection/bypass port 
		]
		pl/stop: ctx/stop-at
	]
	
	open-dns: func [host local-ctx callback /local dns][
		dns: open/no-wait [
			scheme: 'dns
			host: "/async"
			locals: local-ctx
			awake: :callback
		]
		append spwl dns
		insert dns host
		dns
	]
	
	;--- Async I/O handlers ---
	
	would-block?: func [err-obj][err-obj/code = 517]

	on-dns-data: func [dns /local result ctx][
		result: copy dns
		close dns
		remove find spwl dns
		ctx: context append copy [
			locals: context [
				handler: select protocols dns/locals/1
				stop: none			
				events: extract/index reduce dns/locals/2 2 2
				event-names: bind [on-resolved] in handler 'self
			]
			user-data: none
		] any [dns/locals/3 []]
		
		fire-event/arg ctx 'on-received result
		false
	]
	
	on-resolve: func [dns /local port][
		port: dns/locals
		port/remote-ip: copy dns
		close dns
		remove find spwl dns
		either port/remote-ip [
			append spwl port
			if verbose > 1 [log/info reform ["Connecting to" port/host ":" port/port-id]]
			port/async-modes: 'connect
			port/awake: :on-connect
			port/awake port		; try to connect at once
		][
			if verbose > 1 [log/info join "Unknown host: " port/host]
			do select reduce port/user-data 'on-error 'unknown-host port
		]
		false
	]
	
	on-connect: func [port /local err][
		either error? err: try [
			port: open/direct/binary/no-wait port
		][
			either would-block? disarm err [
				if verbose > 2 [
					log/info "opening connection failed...retrying"
				]
				poke port/locals 2 port/locals/2 - 1
				if zero? port/locals/2 [
					if verbose > 1 [log/info "Host unreachable!"]
					remove find spwl port
					do select reduce port/user-data 'on-error 'connect-failed port
				]
			][
				if verbose > 1 [log/info "Host unreachable!"]
				do select reduce port/user-data 'on-error 'unreachable port
			]
		][
			if verbose > 1 [
				log/info reform ["Connected to" port/remote-ip "port:" port/remote-port]
			]
			init-connection port
			fire-event port 'on-connected
			port/async-modes: 'read
			port/awake: :on-data
			on-data port
		]
		false
	]
		
	on-accept: func [server-port /local err new handler][
		either error? err: try [
			new: first server-port 
		][
			if not would-block? disarm err [
				log/error "Async Accept phase failed"
			]
		][
			append spwl new
			if verbose > 0 [
				log/info ["new client: " new/remote-ip " - " server-port/port-id]
			]
			init-connection/service new server-port
			fire-event new 'on-new-client
			new/awake: :on-data
			new/async-modes: [read write]
		]
		false
	]

	on-data: func [port /local in-buf pl state len cut q obj tmp][
		
;--- READ handler ---		
		clear ip-buffer
		either error? err: try [
			state: read-io port ip-buffer ip-buf-size
		][
			if not would-block? disarm err [
				log/error "Async Read phase failed"
				;return true
			]
		][	
			either state <= 0 [
				if verbose > 0 [log/info ["Connection closed by peer " port/remote-ip]]
				close-connection port
			][
				if verbose > 1 [log/info [">> low-level reading: " state]]
				pl: port/locals
				in-buf: pl/in-buffer
				insert tail in-buf ip-buffer
				either pl/stop [
					while [
						any [
							all [
								integer? pl/stop
								pl/stop <= length? in-buf
								cut: pl/stop
							]
							all [
								any [string? pl/stop char? pl/stop]
								cut: find/tail in-buf pl/stop
							]
						]
					][
						fire-event/arg port 'on-received copy/part in-buf cut
						remove/part in-buf cut
					]
				][
					fire-event/arg port 'on-raw-received in-buf
					clear in-buf
				]
			]
		]
;--- end of READ handler ---

;--- WRITE handler ---		
		pl: port/locals
		either empty? q: pl/write-queue [
			if block? port/async-modes [
				port/async-modes: 'read
			]
		][
			if not block? port/async-modes [
				port/async-modes: [read write]
			]
			if file? q/1 [
				change q data: open/read/binary/direct first q
				either data: copy/part first q pl/file-chunk [
					insert q data
				][
					log/error "cannot read file : file empty?. Emission cancelled"
					remove-from-queue q
				]
			]
			either error? err: try [
				state: write-io port first q len: length? first q
			][
				if not would-block? obj: disarm err [
					either obj/code = 501 [						;-- client port closed!
						if verbose > 0 [log/warn "write failed: client port closed"]
						close-connection port
					][
						log/error "Async Write phase failed"
						;return true
					]
				]
			][
				if verbose > 1 [log/info ["<< low-level writing: " state]]
				if state <> -1 [
					either state < -1 [
						if verbose > 0 [log/warn ["write error code " state]]

						close-connection port
					][
						change q skip first q state
						if tail? first q [
							remove-from-queue q
							if port? q/1 [
								either data: copy/part first q pl/file-chunk [
									insert q data
								][
									remove-from-queue q
								]
							]
							if all [pl/flag-close tail? q][
								close-connection port
							]
						]
					]
				]
			]
		]		
;--- end of WRITE handler ---

		false
	]	
	
	control: func [
		name id /start /stop /all /list /install file 
		/local svc p out
	][
		if start [
			either all [
				foreach [name svc] services [
					if not find-service name svc/port-id [
						open-service name svc/port-id
					]
				]
			][
				if not svc: select services name [
					return false
				]
				if none? id [id: get in svc 'port-id]
				if not find-service name id [
					out: open-service name id
				]
				return to-logic out
			]
		]
		if stop [return to-logic close-service name id]
		if list [
			out: make block! 32
			foreach [name svc] services [
				if not find-service name svc/port-id [
					repend out [svc/port-id name svc/module no]
				]
			]
			foreach p spwl [
				if system/words/all [port? p in p 'server-type][
					repend out [
						p/port-id
						p/server-type/name
						p/server-type/module
						yes
					]
				]
			]
			return out
		]
		true
	]
	
	boot: func [/no-loop /with spec [block!] /local name file svc init path][
		if value? 'uniserve-path [
			path: what-dir
			change-dir uniserve-path
		]		
		if exists? services-path [
			foreach file read services-path/. [
				if script? svc: services-path/:file [
					either with [
						name: to-word to-string copy/part file find file #"."
						if all [find spec 'services find spec/services name][do svc]
					][do svc]
				]
			]
		]
		if exists? protocols-path [
			foreach file read protocols-path/. [
				if script? svc: protocols-path/:file [
					either with [
						name: to-word to-string copy/part file find file #"."
						if all [find spec 'protocols find spec/protocols name][do svc]
					][do svc]
				]
			]
		]
		foreach [name svc] services [
			svc/shared: uniserve/shared
			if all [init: in svc 'on-load function? get init][do init]
			control/start name none
			log/info ["starting " mold name "..."]
		]
		if path [change-dir path]
		if not no-loop [wait []]
	]

	install-plugin: func [
		body [block!]
		/as-service
		/local new pos class callback list
	][
		set [class callback list] reduce any [
			all [as-service [service-class 'on-new-client services]]
			[protocol-class 'on-connected protocols]
		]
		new: make log-class []
		new: make class new
		new: make new body

		either any [
			get in new 'on-received
			all [not new/stop-at get in new 'on-raw-received]
			all [not new/stop-at get in new callback]
		][
			either pos: find list new/name [
				change pos/2 new
			][
				repend list [new/name new]
			]
		][
			log/error rejoin ["no valid receive callback in " mold new/name ":" new/port-id]
		]
		new
	]

	set 'install-service  func [body][install-plugin/as-service body]
	
	set 'install-protocol func [body /local plugin new evt evt-names name value][
		foreach evt evt-names: select body to-set-word 'events [
			append body to-set-word evt
		]
		append body none
		plugin: install-plugin body
		log/info ["Async Protocol " mold plugin/name " loaded"]
	]
	
	shared: context [control: none]
	shared/control: :control
]

;-- last REBOL versions load IE proxy parameters automatically
;-- we reset the proxy host to avoid side effects.
system/schemes/default/proxy/host:
system/schemes/default/proxy/port-id: none

system/options/quiet: true
